from .specs import Pipeline, Workflow
from .queues import make_queues, MakeQueue, PipelineQueues, WorkflowQueues

__all__ = [
  'Pipeline', 'Workflow',
  'make_queues', 'MakeQueue', 'PipelineQueues', 'WorkflowQueues',
]


# BEGIN
from typing import Unpack, TypedDict
from dataclasses import dataclass
from q.api import WriteQueue
from pipeteer import MakeQueue, make_queues as _make_queues, PipelineQueues
# UNCOMMENT from MODULE import MOD_IMPORTS

class SUBFLOW_CLASS: # DELETE
  Queues = int # DELETE
# DELETE
INPUTS = str; OUTPUTS = str; INPUT_TYPE = str; OUTPUT_TYPE = str # DELETE
class WORKFLOW:
  class Queues(TypedDict):
    Qin: WriteQueue[INPUT_TYPE]
    # LOOP SUBFLOW_ID SUBFLOW_CLASS
    SUBFLOW_ID: SUBFLOW_CLASS.Queues
    # END
    # LOOP PIPELINE_ID INPUTS OUTPUTS
    PIPELINE_ID: PipelineQueues[INPUTS, OUTPUTS]
    # END

  @staticmethod
  def make_queues(make_queue: MakeQueue, output_queue: WriteQueue[OUTPUT_TYPE]) -> Queues:
    VARIABLE = ... # type: ignore # DELETE
    return _make_queues(VARIABLE, make_queue, output_queue) # type: ignore
  
  @staticmethod
  def artifacts(**queues: Unpack['WORKFLOW.Queues']):
    ...
    
# END
from typing import Sequence
from haskellian import Thunk
from templang import parse
from pipeteer import Workflow

source = Thunk(lambda: open(__file__).read())

def union_type(types: Sequence[type]):
  return  ' | '.join(t.__name__ for t in types)

def codegen(workflow: Workflow, *, variable: str, module: str, classname: str) -> str:

  translations = {
    'MODULE': module,
    'WORKFLOW': classname,
    'VARIABLE': variable,
    'INPUT_TYPE': workflow.input_type.__name__,
    'OUTPUT_TYPE': union_type(workflow.output_types),
    'MOD_IMPORTS': '*',
    'SUBFLOW_CLASS': [],
    'SUBFLOW_ID': [],
    'PIPELINE_CLASS': [],
    'PIPELINE_ID': [],
    'INPUTS': [],
    'OUTPUTS': [],
  }

  for id, pipe in workflow.pipelines.items():
    match pipe:
      case Workflow():
        translations['SUBFLOW_CLASS'].append(id.title())
        translations['SUBFLOW_ID'].append(id)
      case _:
        translations['INPUTS'].append(pipe.input_type.__name__)
        translations['OUTPUTS'].append(union_type(pipe.output_types))
        translations['PIPELINE_CLASS'].append(id.title())
        translations['PIPELINE_ID'].append(id)

  return parse(source(), translations)
  


from typing import Mapping, Protocol, TypedDict, Generic, Sequence, TypeVar, TypeAlias
from haskellian import Left, Either
from q.api import Queue, ReadQueue, WriteQueue, QueueError
from .specs import Pipeline, Workflow

A = TypeVar('A')
B = TypeVar('B')

class PipelineQueues(TypedDict, Generic[A, B]):
  Qin: ReadQueue[A]
  Qout: WriteQueue[B]

WorkflowQueues: TypeAlias = PipelineQueues | Mapping[str, 'WorkflowQueues']

class prejoin(WriteQueue[A], Generic[A]):
  def __init__(self, queues: Sequence[tuple[WriteQueue[A], Sequence[type[A]]]]):
    self.queues = queues

  async def push(self, key: str, value: A) -> Either[QueueError, None]:
    for q, types in self.queues:
      for t in types:
        if isinstance(value, t):
          return await q.push(key, value)
    return Left(QueueError(f'Invalid type {type(value).__name__} for {value}'))

class MakeQueue(Protocol):
  def __call__(self, id: Sequence[str], type: type[A]) -> Queue[A]:
    ...

def make_queues(
  workflow: Workflow,
  make_queue: MakeQueue,
  output_queue: WriteQueue
) -> WorkflowQueues:
  
  def _input_queue(task: Pipeline, prefix: tuple[str, ...]) -> tuple[Queue[A], Sequence[type[A]]]:
    match task:
      case Workflow():
        return _input_queue(task.pipelines[task.input_task], prefix + (task.input_task,))
      case Pipeline():
        return make_queue(prefix, task.input_type), [task.input_type]
      
  def _make_queues(task: Pipeline, prefix: tuple[str, ...], output_queue: WriteQueue) -> WorkflowQueues:
    match task:
      case Workflow():
        queues = {
          id: _make_queues(
            pipe, prefix + (id,),
            prejoin([
              ((output_queue, task.output_types) if id == 'output' else _input_queue(task.pipelines[id], prefix + (id,)))
              for id in task.next_tasks(pipe.output_types)
            ])
          )
          for id, pipe in task.pipelines.items()
        }
        return queues | { 'Qin': queues[task.input_task]['Qin'] } # type: ignore
      case Pipeline():
        return PipelineQueues(
          Qin=_input_queue(task, prefix)[0],
          Qout=output_queue
        )

  return _make_queues(workflow, (), output_queue)
